@@ -44,10 +44,11 @@
 
 #ifndef MAX
 #define MAX(x, y) ((x) > (y) ? (x) : (y))
+#define MIN(x, y) ((x) < (y) ? (x) : (y))
 #endif
 
 #ifdef PTHREAD_STACK_MIN
-#define THREAD_STACK_SIZE MAX(8*1024, PTHREAD_STACK_MIN)
+#define THREAD_STACK_SIZE MAX(16*1024, PTHREAD_STACK_MIN)
 #else
 #define THREAD_STACK_SIZE 64*1024
 #endif
@@ -112,8 +113,8 @@ struct thread {
 static void dolog(const char* fmt, ...) { }
 #endif
 
-static struct addrinfo* addr_choose(struct addrinfo* list, union sockaddr_union* bind_addr) {
-	int af = SOCKADDR_UNION_AF(bind_addr);
+static struct addrinfo* addr_choose(struct addrinfo* list, union sockaddr_union* bindaddr) {
+	int af = SOCKADDR_UNION_AF(bindaddr);
 	if(af == AF_UNSPEC) return list;
 	struct addrinfo* p;
 	for(p=list; p; p=p->ai_next)
@@ -282,7 +283,10 @@ static void copyloop(int fd1, int fd2) {
 		}
 		int infd = (fds[0].revents & POLLIN) ? fd1 : fd2;
 		int outfd = infd == fd2 ? fd1 : fd2;
-		char buf[1024];
+		/* since the biggest stack consumer in the entire code is
+		   libc's getaddrinfo(), we can safely use at least half the
+		   available stacksize to improve throughput. */
+		char buf[MIN(16*1024, THREAD_STACK_SIZE/2)];
 		ssize_t sent = 0, n = read(infd, buf, sizeof buf);
 		if(n <= 0) return;
 		while(sent < n) {
@@ -310,14 +314,12 @@ static enum errorcode check_credentials(unsigned char* buf, size_t n) {
 	return EC_NOT_ALLOWED;
 }
 
-static void* clientthread(void *data) {
-	struct thread *t = data;
-	t->state = SS_1_CONNECTED;
+static int handshake(struct thread *t) {
 	unsigned char buf[1024];
 	ssize_t n;
 	int ret;
-	int remotefd = -1;
 	enum authmethod am;
+	t->state = SS_1_CONNECTED;
 	while((n = recv(t->client.fd, buf, sizeof buf, 0)) > 0) {
 		switch(t->state) {
 			case SS_1_CONNECTED:
@@ -325,13 +327,13 @@ static void* clientthread(void *data) {
 				if(am == AM_NO_AUTH) t->state = SS_3_AUTHED;
 				else if (am == AM_USERNAME) t->state = SS_2_NEED_AUTH;
 				send_auth_response(t->client.fd, 5, am);
-				if(am == AM_INVALID) goto breakloop;
+				if(am == AM_INVALID) return -1;
 				break;
 			case SS_2_NEED_AUTH:
 				ret = check_credentials(buf, n);
 				send_auth_response(t->client.fd, 1, ret);
 				if(ret != EC_SUCCESS)
-					goto breakloop;
+					return -1;
 				t->state = SS_3_AUTHED;
 				if(auth_ips && !pthread_rwlock_wrlock(&auth_ips_lock)) {
 					if(!is_in_authed_list(&t->client.addr))
@@ -343,23 +345,24 @@ static void* clientthread(void *data) {
 				ret = connect_socks_target(buf, n, &t->client);
 				if(ret < 0) {
 					send_error(t->client.fd, ret*-1);
-					goto breakloop;
+					return -1;
 				}
-				remotefd = ret;
 				send_error(t->client.fd, EC_SUCCESS);
-				copyloop(t->client.fd, remotefd);
-				goto breakloop;
-
+				return ret;
 		}
 	}
-breakloop:
+	return -1;
+}
 
-	if(remotefd != -1)
+static void* clientthread(void *data) {
+	struct thread *t = data;
+	int remotefd = handshake(t);
+	if(remotefd != -1) {
+		copyloop(t->client.fd, remotefd);
 		close(remotefd);
-
+	}
 	close(t->client.fd);
 	t->done = 1;
-
 	return 0;
 }
 
@@ -380,17 +383,21 @@ static int usage(void) {
 	dprintf(2,
 		"MicroSocks SOCKS5 Server\n"
 		"------------------------\n"
-		"usage: microsocks -1 -q -i listenip -p port -u user -P password -b bindaddr\n"
+		"usage: microsocks -1 -q -i listenip -p port -u user -P pass -b bindaddr -w ips\n"
 		"all arguments are optional.\n"
 		"by default listenip is 0.0.0.0 and port 1080.\n\n"
 		"option -q disables logging.\n"
 		"option -b specifies which ip outgoing connections are bound to\n"
+		"option -w allows to specify a comma-separated whitelist of ip addresses,\n"
+		" that may use the proxy without user/pass authentication.\n"
+		" e.g. -w 127.0.0.1,192.168.1.1.1,::1 or just -w 10.0.0.1\n"
+		" to allow access ONLY to those ips, choose an impossible to guess user/pw combo.\n"
 		"option -1 activates auth_once mode: once a specific ip address\n"
-		"authed successfully with user/pass, it is added to a whitelist\n"
-		"and may use the proxy without auth.\n"
-		"this is handy for programs like firefox that don't support\n"
-		"user/pass auth. for it to work you'd basically make one connection\n"
-		"with another program that supports it, and then you can use firefox too.\n"
+		" authed successfully with user/pass, it is added to a whitelist\n"
+		" and may use the proxy without auth.\n"
+		" this is handy for programs like firefox that don't support\n"
+		" user/pass auth. for it to work you'd basically make one connection\n"
+		" with another program that supports it, and then you can use firefox too.\n"
 	);
 	return 1;
 }
@@ -404,11 +411,27 @@ static void zero_arg(char *s) {
 int main(int argc, char** argv) {
 	int ch;
 	const char *listenip = "0.0.0.0";
+	char *p, *q;
 	unsigned port = 1080;
-	while((ch = getopt(argc, argv, ":1qb:i:p:u:P:")) != -1) {
+	while((ch = getopt(argc, argv, ":1qb:i:p:u:P:w:")) != -1) {
 		switch(ch) {
+			case 'w': /* fall-through */
 			case '1':
-				auth_ips = sblist_new(sizeof(union sockaddr_union), 8);
+				if(!auth_ips)
+					auth_ips = sblist_new(sizeof(union sockaddr_union), 8);
+				if(ch == '1') break;
+				p = optarg;
+				while(1) {
+					union sockaddr_union ca;
+					if((q = strchr(p, ','))) *q = 0;
+					if(resolve_sa(p, 0, &ca)) {
+						dprintf(2, "error: failed to resolve %s\n", p);
+						return 1;
+					}
+					add_auth_ip(&ca);
+					if(q) *(q++) = ',', p = q;
+					else break;
+				}
 				break;
 			case 'q':
 				quiet = 1;
@@ -442,7 +465,7 @@ int main(int argc, char** argv) {
 		return 1;
 	}
 	if(auth_ips && !auth_pass) {
-		dprintf(2, "error: auth-once option must be used together with user/pass\n");
+		dprintf(2, "error: -1/-w options must be used together with user/pass\n");
 		return 1;
 	}
 	signal(SIGPIPE, SIG_IGN);
